"""
Commands to load a configuration
"""
import imp
import json
import os
import re
import sys
import uuid

from .attrs import attrs
from .validation import validate, is_spec, ValidationError
from .utils import normalize_keys


def load_config(path, relative_to_caller=False):
  """Load a config module as a dict

  Parameters
  ----------
  path : str
      path to configuration file
  relative_to_caller : bool
      if path isn't an absolute file path, interpret it as relative to the file
      from which this method is called.
  """
  if relative_to_caller and not os.path.abspath(path) == path:
    previous_fname  = previous_frame().f_globals['__file__']
    previous_fname  = os.path.abspath(previous_fname)
    previous_folder = os.path.split(previous_fname)[0]
    path            = os.path.join(previous_folder, path)
  modname = str(uuid.uuid4())
  module = imp.load_source(modname, path)
  variables = {
      k:getattr(module, k) for k in dir(module)
      if not '__' in k
  }
  return normalize_keys(variables)


def load_spec(path, relative_to_caller=False):
  """Load a configuration specification"""
  spec = load_config(path, relative_to_caller=relative_to_caller)
  spec = { k:v for k,v in spec.items()
           if is_spec(v) }
  return spec


def import_config(path, relative_to_caller=True):
  """Load contents of another configuration file into the environment

  Parameters
  ----------
  path : str
      path to configuration file
  relative_to_caller : bool
      if path isn't an absolute file path, interpret it as relative to the file
      from which this method is called.
  """
  variables = load_config(path, relative_to_caller=relative_to_caller)
  _add_globals(**variables)


def import_spec(path, relative_to_caller=True):
  """Import another specification's contents into the environment"""
  variables = load_spec(path, relative_to_caller=relative_to_caller)
  _add_globals(**variables)


def env(variable_name):
  """Load an environment variable

  Parameters
  ----------
  variable_name : str
      environment variable name
  """
  return os.environ[variable_name]


def update_config(old, new):
  def _split(k):
    # .a.b.c -> (a, b.c)
    # .1.b.c -> (1, b.c)
    # [1].b[2].c -> (1, .b[2].c)
    # .a -> (a, '')
    # [1] -> (1, '')
    if len(k) == 0:
      return []
    elif k.startswith('['):
      match = re.search(r'^\[([-+]?\d+?)\]', k)
      start = int(match.group(1))
      rest  = k[match.end():]
      rest = _split(rest)
      return [start] + rest
    elif k.startswith('.'):
      match = re.search(r'[.](.+?)(\[|[.]|$)', k)
      start = match.group(1)
      if match.end() == len(k):
        rest = k[match.end():]
      else:
        rest = k[match.end()-1:]
      rest = _split(rest)
      try:
        start = int(start)
      except ValueError:
        pass
      return [start] + rest
    else:
      wtf

  def _reconstruct(keys, v):
    if len(keys) == 0:
      return v
    else:
      start, rest = keys[0], keys[1:]
      if isinstance(start, int):
        raise ValueError("Unable to reconstruct command line argument with array indices")
      else:
        return {start: _reconstruct(rest, v)}

  def _update(o, keys, v):
    if len(keys) == 0:
      return v
    else:
      if o is None:
        return _reconstruct(keys, v)
      else:
        start, rest = keys[0], keys[1:]
        if isinstance(start, int):
          if isinstance(o, list):
            o[start] = _update(o[start], rest, v)
          elif isinstance(o, tuple):
            o = list(o)
            o[start] = _update(o[start], rest, v)
            return tuple(o)
        elif isinstance(start, basestring):
          o[start] = _update(o.get(start, None), rest, v)
        else:
          wtf
        return o

  for (k, v) in new.items():
    old = _update(old, _split('.' + k), v)
  return old


def configure(args=None, config_path=None, spec_path=None):
  """Initialize configuration"""
  # load configuration file, if one was specified
  if config_path is not None:
    config = load_config(config_path)
  else:
    config = {}

  # override config file with command line args
  args   = _eat_command_line_arguments(args)
  config = update_config(config, args)

  # validate the config
  if spec_path:
    spec   = load_spec(spec_path)
    config = validate(spec, config)

  return attrs.from_dict(config)


def _eat_command_line_arguments(args=None):
  """Turn command line args into key-value pairs

  Assume command line args are of the form
    --key1 value1 --key2 value2 ...

  values must be valid python expressions
  """
  result = {}
  if args is None:
    args = sys.argv[1:]
  i = 0
  while i < len(args):
    assert args[i].startswith('--'), \
        'Command line options must be of the form "--key value'
    key = args[i][2:]
    try:
      value = eval(args[i+1])
    except NameError:
      value = args[i+1]
    result[key] = value
    i += 2
  return normalize_keys(result)


def _add_globals(**kwargs):
  """Update global dictionary of whoever is using this module"""
  # get first stack frame not in this module
  # update environment of said frame
  glob = previous_frame().f_globals
  glob.update(kwargs)


def previous_frame():
  """Find the first frame in the call stack not originating from this module"""
  i = 0
  while sys._getframe(i).f_globals['__name__'] == __name__:
    i += 1
  return sys._getframe(i)



import re


def identity(x):
  return x


def recursive_apply(obj, key_func=identity, value_func=identity):
  """Recursively apply a function to the keys and values of a nested object"""
  if isinstance(obj, dict):
    result = {}
    for k, v in obj.items():
      k = key_func(k)
      result[k] = recursive_apply(v, key_func, value_func)
    return result
  elif isinstance(obj, list):
    result = []
    for v in obj:
      result.append(recursive_apply(v, key_func, value_func))
    return result
  elif isinstance(obj, tuple):
    result = []
    for v in obj:
      result.append(recursive_apply(v, key_func, value_func))
    return tuple(result)
  else:
    return value_func(obj)


def normalize_keys(obj):
  """replace list-variable with list_variable"""

  def key_func(k):
    return re.sub("([a-zA-Z])-([a-zA-Z])", r"\1_\2", k)

  return recursive_apply(obj, key_func=key_func)



from .commands import configure, import_config, load_config, load_spec, import_spec, env
from .validation import required, optional, one_of, ValidationError


__all__ = [
    'configure',
    'import_config',
    'load_config',
    'import_spec',
    'load_spec',
    'optional',
    'required',
    'env',
    'one_of',
    'ValidationError',
]



"""
Tools for validating a configuration spec
"""

from utils import identity


class variable(object):
  def __init__(self, type=str, help=None):
    self.type = type
    self.help = help

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    return isinstance(other, variable) and \
        other.type == self.type and \
        other.help == self.help


class optional(variable):
  """An optional variable

  Used in a spec file to denote a variable that has a default value and may be
  left unspecified in a config file.

  >>> version = optional(default="1.0-SNAPSHOT", \
                         help="Version of this application")

  Parameters
  ----------
  type : function
      function to coerce argument to desired type
  help : str
      long description of this variable
  default : object
      default value to use if this variable isn't in the config.
  """
  def __init__(self, type=identity, help=None, default=None):
    super(optional, self).__init__(type, help)
    self.default = default

  def __str__(self):
    return (
        "optional(type=%s, help=%s, default=%s)" %
        (self.type, self.help, self.default)
    )

  def __eq__(self, other):
    return super(optional, self).__eq__(other) and \
        isinstance(other, optional) and \
        other.default == self.default


class required(variable):
  """A required variable

  Used in a spec file to denote a variable that must be in a valid config file.
  If a required variable is not in a config file, validation will fail.

  Parameters
  ----------
  type : function
      function to coerce argument to desired type
  help : str
      long description of this variable
  """
  def __init__(self, type=identity, help=None):
    super(required, self).__init__(type, help)

  def __str__(self):
    return (
        "required(type=%s, help=%s)" %
        (self.type, self.help)
    )

  def __eq__(self, other):
    return super(required, self).__eq__(other) and \
        isinstance(other, required)


def is_required(obj):
  """Is this a required variable?"""
  return is_collection(obj) \
      or isinstance(obj, required)


def is_spec(obj):
  """Is this a valid specification variable?"""
  return is_collection(obj) \
      or isinstance(obj, variable)


def is_collection(obj):
  return isinstance(obj, dict) \
      or isinstance(obj, list) \
      or isinstance(obj, tuple)


class ValidationError(Exception):
  def __init__(self, msg, path=''):
    super(ValidationError, self).__init__("%s (%s)" % (msg, path[1:]))
    self.msg  = msg
    self.path = path

  def extend_path(self, path):
    return ValidationError(path=(path + self.path), msg=self.msg)


def validate(spec, config):
  """Validate a configuration file against a spec"""
  if is_required(spec) and config is None:
    raise ValidationError("Missing required argument")

  # unpack optional collection
  if isinstance(spec, variable) and is_collection(spec.type):
    if isinstance(spec, optional):
      if config is None:
        return spec.default
      else:
        spec = spec.type
    elif isinstance(spec, required):
      spec = spec.type
    else:
      raise ValidationError(msg="Unrecognized variable type: %s" % str(type(spec)))

  if isinstance(spec, dict):
    if not isinstance(config, dict):
      raise ValidationError("Improper object type (expected dict)")
    result = {}
    for k, v in spec.items():
      try:
        result[k] = validate(v, config.get(k, None))
      except ValidationError as e:
        raise e.extend_path(".%s" % k)
    return result

  if isinstance(spec, list):
    if not isinstance(config, list):
      raise ValidationError("Improper object type (expected list)")
    result = []
    for i, v in enumerate(config):
      try:
        result.append(validate(spec[0], v))
      except ValidationError as e:
        raise e.extend_path("[%d]" % i)
    return result

  if isinstance(spec, tuple):
    if not isinstance(config, tuple):
      raise ValidationError("Improper object type (expected tuple)")
    if not len(spec) == len(config):
      raise ValidationError("Incorrect tuple length (expected %d; found %d)" % \
          (len(spec), len(config)))
    result = []
    for i, (s, v) in enumerate(zip(spec,config)):
      try:
        result.append(validate(s, v))
      except ValidationError as e:
        raise e.extend_path("[%d]" % i)
    return tuple(result)

  if isinstance(spec, required):
    if config is None:
      raise ValidationError("Missing required argument")
    else:
      try:
        return spec.type(config)
      except ValueError as e:
        raise ValidationError("Type conversion failure")

  if isinstance(spec, optional):
    if config is None:
      return spec.default
    else:
      try:
        return spec.type(config)
      except ValueError as e:
        raise ValidationError("Type conversion failure")


def one_of(**options):
  """is this object one of these options?"""
  def type(obj):
    if not obj in options:
      raise ValueError("%s isn't one of %s" % (obj, options))
    return obj
  return type



from setuptools import setup, find_packages

setup(
    name         = 'configurati',
    version      = '0.1.1',
    author       = 'Daniel Duckworth',
    author_email = 'pykalman@gmail.com',
    description  = 'Configuration files for Python',
    license      = 'BSD',
    keywords     = 'config configuration',
    url          = 'github.com/duckworthd/configurati',
    packages     = find_packages(),
    classifiers  = [
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: BSD License',
      'Operating System :: OS Independent',
      'Programming Language :: Python',
      'Programming Language :: Python :: Implementation :: CPython',
      'Topic :: Software Development :: Libraries',
    ],
    install_requires = [],
    tests_require    = [
      'nose',
    ]
)



"""
Attribute dictionary
"""

class attrs(dict):
  """A dict you can access with dot notation"""

  def __getattr__(self, key):
    return self[key]

  @staticmethod
  def from_dict(d):
    if isinstance(d, dict):
      result = attrs( (k, attrs.from_dict(v)) for (k, v) in d.items() )
    elif isinstance(d, list):
      result = [attrs.from_dict(v) for v in d]
    elif isinstance(d, tuple):
      result = tuple([attrs.from_dict(v) for v in d])
    else:
      result = d
    return result
